from pydantic import BaseModel
from acex.models.attribute_value import AttributeValue
from enum import Enum
from typing import Optional, Dict


class LoggingServerBase(BaseModel): ...
    #name: str = None


class LoggingSeverity(str, Enum): 
    EMERGENCY = "EMERGENCY"
    ALERT = "ALERT"
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    NOTICE = "NOTICE"
    INFORMATIONAL = "INFORMATIONAL"
    DEBUG = "DEBUG"


class LoggingConfig(BaseModel):
    rate_limit: Optional[AttributeValue[int]] = None
    severity: Optional[AttributeValue[LoggingSeverity]] = None

class LoggingConsole(BaseModel):
    name: str = None
    line_number: int = None
    logging_synchronous: bool = True

class RemoteServer(BaseModel):
    name: str = None
    host: str = None
    port: int = 514
    transfer: str = 'udp'
    source_address: str = None

class VtyLines(BaseModel):
    name: str = None
    line_number: int = None
    logging_synchronous: bool = True

class GlobalConfig(BaseModel):
    name: str = None
    buffer_size: int = 4096

class FileConfig(BaseModel):
    name: str = None # object name
    filename: str = None # name of the file
    rotate: int = None # How many versions to keep
    max_size: int = None # Think Ciscos "logging buffered"
    facility: str = None # Level for logs


class LoggingEvent(BaseModel):
    enabled: bool
    severity: LoggingSeverity


class LoggingEvents(BaseModel):
    events: Optional[Dict[str, LoggingEvent]] = {}



from .integration_plugin_base import IntegrationPluginBase

class DatasourceInMemory(IntegrationPluginBase): 

    def __init__(self):
        self.db = {}

    def create(self, item: dict): 
        _id = str(len(self.db)+1)
        item = dict(item)
        item["id"] = _id
        self.db[str(len(self.db)+1)] = item
        return True

    def get(self, id: str): 
        return self.db.get(id)

    def query(self): 
        items = []
        for k, v in self.db.items():
            items.append(v)
        return items

    def update(self,id: str, data: dict ): 
        for k, v in data.items():
            self.db[id][k] = v

    def delete(self, id: str): 
        self.db.pop(id)


from sqlmodel import SQLModel, Field
from typing import Optional, Dict
from enum import Enum


class AssetRefType(str, Enum):
    asset = "asset"
    assetcluster = "assetcluster"


class NodeBase(SQLModel):
    asset_ref_id: int
    asset_ref_type: AssetRefType = AssetRefType.asset
    logical_node_id: int

class Node(NodeBase, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)

class NodeResponse(NodeBase):
    asset: Dict = Field(default_factory=dict)
    logical_node: Dict = Field(default_factory=dict)



from collections import defaultdict
from ipaddress import IPv4Interface, IPv6Interface, IPv4Address
from pydantic import BaseModel
import json, hashlib
from typing import Dict, Any, Type, Union, Optional, get_origin
from types import NoneType
from datetime import datetime
from acex.models import ExternalValue, AttributeValue


class ConfigComponent:
    type: str = "component"
    name: str = None
    model_cls: Type[BaseModel] = None

    def __init__(self, *args, **kwargs):
        self.kwargs = kwargs

        # Hook for preprocessing kwargs before initialization
        if hasattr(self, "pre_init"):
            getattr(self, "pre_init")()

        # Validate against the model for the component type
        self.model = self._validate_model(self.kwargs)

        # Set name for component, must always be unique.
        # For single attribute values, name is same as the single positional arg.
        self._set_name_attribute()


    def _validate_model(self, kwargs) -> BaseModel:
        """
        Validate all kwargs against the model and set attribute
        types accordingly
        """
        if not self.__class__.model_cls:
            raise ValueError(f"No model_cls defined for {self.__class__.__name__}")
        try:
            # Create an instance of the model class with kwargs
            model_instance = self.__class__.model_cls(**kwargs)
            return model_instance
        except Exception as e:
            raise ValueError(f"Failed to validate kwargs against model {self.__class__.model_cls.__name__}: {e}")


    def _set_name_attribute(self):
        """
        Set name attribute to component, not included in the model
        but will have to be unique for mapping/dict-key in the composite configuration.

        # If self.model is AttributeValue[str], that means its a "single-attr 
        component which has no requirement for name, since it can only be one 
        of them in the config. So its name is same as the class name.
        """
        if isinstance(self.model, AttributeValue[str]):
            self.name = self.__class__.__name__.lower()
        else:
            self.name = self.kwargs.get("name")





from pydantic import BaseModel
from acex.configuration.components import ConfigComponent
from acex.configuration.components.interfaces import (
    Loopback,
    Physical,
    Subinterface,
    Svi
)
from acex.configuration.components.system import (
    HostName,
    Contact,
    Location,
    DomainName
)

from acex.configuration.components.system.ntp import NtpServer
from acex.configuration.components.system.ssh import SshServer, AuthorizedKey
from acex.configuration.components.network_instances import NetworkInstance, L3Vrf
from acex.configuration.components.vlan import Vlan
from acex.models.attribute_value import AttributeValue

from acex.models import ExternalValue
from acex.models.composed_configuration import ComposedConfiguration, ReferenceTo, ReferenceFrom, RenderedReference
from collections import defaultdict
from typing import Dict
from string import Template
import json


class Configuration: 
    # Mapping from component type to path in composed configuration
    # Note that some paths are containers, like interfaces where the component also
    # must be referenced using its name attribute
    COMPONENT_MAPPING = {
        AuthorizedKey: "system.ssh.host_keys",
        HostName: "system.config.hostname",
        Contact: "system.config.contact",
        Location: "system.config.location",
        DomainName: "system.config.domain_name",
        Loopback: "interfaces", 
        Physical: "interfaces",
        Subinterface: "interfaces",
        NetworkInstance: "network_instances",
        L3Vrf: "network_instances",
        Vlan: Template("network_instances.${network_instance}.vlans"),
        Svi: Template("interfaces"),
        NtpServer: "system.ntp.servers",
        SshServer: "system.ssh.config"
    }

    # Reverse mapping from attribute name to path for __getattr__
    ATTRIBUTE_TO_PATH = {
        "hostname": "system.config.hostname",
        "contact": "system.config.contact",
        "location": "system.config.location",
        "domain_name": "system.config.domain_name",
    }
    
    def __init__(self, logical_node_id):
        self.composed = ComposedConfiguration()
        self.logical_node_id = logical_node_id

        # Komponenter lagras som objekt, mappade till sin position
        self._components = []

        # Lagra alla Reference object. Läggs till efter att komponenter lagts till
        self._references = []

    
    def _get_nested_component(self, path: str):
        """Get a nested attribute using dot notation path."""
        obj = self.composed
        parts = path.split('.')
        for part in parts:
            obj = getattr(obj, part)
        return obj

    def _set_attr_ptr_on_attributes(self, component):
        """
        Set attr_ptr metadata on externalValue attributes of the component.
        # TODO: If singleAttribute class, implementsmart logic to not use key 'value'?
        """
        # example: logical_nodes.0.ethernetCsmacd.if0.ipv4
        if component.name is not None:
            base_path = f"logical_nodes.{self.logical_node_id}.{component.type}.{component.name}"
        else:
            base_path = f"logical_nodes.{self.logical_node_id}.{component.type}"

        for k, v in component.model.model_dump().items():
            attribute_value = getattr(component.model, k)
            if isinstance(attribute_value, AttributeValue):
                if attribute_value is not None and isinstance(attribute_value.value, ExternalValue):
                    attribute_value.metadata["attr_ptr"] = f"{base_path}.{k}"


    def _get_component_path(self, component) -> str:
        mapped_path = self.COMPONENT_MAPPING[type(component)]
        return self._render_path(component, mapped_path)

    def _render_path(self, component, mapped_path: str):
        """
        The mapped path can either be a pure string, 
        or it can be a string.Template. If the latter, we need 
        to extract the variables and fetch corresponding values which 
        is expected to be attributeValues of the component itself.     
        """

        if isinstance(mapped_path, Template):
            vars_needed = [m.group('named') or m.group('braced')
               for m in mapped_path.pattern.finditer(mapped_path.template)
               if m.group('named') or m.group('braced')]

            value_map = {}
            for v in vars_needed:
                if hasattr(component.model, v):
                    av = getattr(component.model, v)
                    value_map[v] = av.value

            mapped_path = mapped_path.substitute(value_map)
        return mapped_path

    def _pop_all_references(self, component):
        """
        Pops all references and stores them in a flat list instead.
        Source of the reference must always be an attribute, destination 
        can be an object. 
        --> self._references
        """
        for k,v in component.kwargs.items():

            if isinstance(v, (ReferenceFrom, ReferenceTo)):
                self_component_path = self.COMPONENT_MAPPING[type(component)]
                rendered_self_component_path = self._render_path(component, self_component_path)

                # use key if item in dict
                if rendered_self_component_path.endswith('.config'):
                    self_path = f"{rendered_self_component_path}"
                else:    
                    self_path = f"{rendered_self_component_path}.{component.name}"

                if isinstance(v, ReferenceTo):
                    # if self is the source, the source must be an attribute. 
                    ri = RenderedReference(
                        from_ptr = f"{self_path}.{k}",
                        to_ptr = v.pointer
                    )
                elif isinstance(v, ReferenceFrom):
                    
                    ri = RenderedReference(
                        from_ptr = v.pointer,
                        to_ptr = f"{self_path}"
                    )
                self._references.append(ri)

                

    def add(self, component):
        """
        Add a configuration component to the composed configuration.
        Handles both dict-based and single-value components, sets external refs, and logs actions.
        Returns the added value or raises ValueError on error.
        """
        component_type = type(component)
        if component_type not in self.COMPONENT_MAPPING:
            print(f"Unknown component type: {component_type.__name__}")
            raise ValueError(f"Unknown component type: {component_type.__name__}")

        # pop references
        self._pop_all_references(component)

        # ref is used to reference the absolute path of each attribute:
        self._set_attr_ptr_on_attributes(component)

        # modellen för composed talar om ifall vi behöver ett key för componenten:
        # Fix composite path for mapped objects. 
        composite_path = self._get_component_path(component)

        if composite_path is not None:
            # lägger till komponenten i en flat list i en tuple med path.
            self._components.append((composite_path, component))

    def as_model(self) -> BaseModel:

        # Dont edit the actual composed model, we make a model from a copy
        config = self.composed.copy()

        # Apply all values from components to the composed model: 
        for path, component in self._components:

            # Set metadata of the component
            if hasattr(component.model, "metadata"):
                component.model.metadata.type = component.type

            # Traverse the composed object to the ptr for the obj.
            path_parts = path.split('.')
            attribute_name = path_parts.pop()

            # First place the pointer on the attribute key
            ptr = config
            for part in path_parts:
                if isinstance(ptr, dict):
                    ptr = ptr.get(part)
                elif hasattr(ptr, part):
                    ptr = getattr(ptr, part)
                else:
                    raise Exception(f"Component pointer invalid for component: {component}, path: {path_parts}")

            # Get value of the pointer
            if isinstance(ptr, dict):
                value = ptr.get(attribute_name)
            else:
                value = getattr(ptr, attribute_name)
            
            # If the value of the ptr is a dict, the item has to be keyed
            if isinstance(value, dict):
                value[component.name] = component.model
            else:
                # Cant set value directly, since NoneType is a singleton.
                # Instead we use setattr using the pointer and attribute name.
                setattr(ptr, attribute_name, component.model)

        # Add and resolve references:
        for reference in self._references:
            # From pointer is always an attribute, to is a referenced object

            # insertion point 
            insertion_path_parts = reference.from_ptr.split('.')
            insertion_attr = insertion_path_parts.pop()

            # referenced value
            value_path_parts = reference.to_ptr.split('.')
            value_attr = value_path_parts.pop()

            # pointer startpoint is the full config.
            ptr = config

            # Move pointer to the value
            for part in value_path_parts:
                if isinstance(ptr, dict):
                    ptr = ptr.get(part)
                else:
                    ptr = getattr(ptr, part)
            
            # Store pointer value, this will be inserted at insertion point
            pointer_value = ptr[value_attr]

            # Reset pointer to base of configuration
            ptr = config

            # Move pointer to insertion point
            for part in insertion_path_parts:
                if isinstance(ptr, dict):
                    ptr = ptr.get(part)
                else:
                    ptr = getattr(ptr, part)

            value = {
                "name": pointer_value.name.value,
                "metadata": {
                    "type": "reference",
                    "ref_path": reference.to_ptr
                }
            }
            # Insert the referenced value dict to the insertion point.
            # If attribute of insertionpoint is a dict, value has to be keyed
            if isinstance(getattr(ptr, insertion_attr), dict):
                insertion_point = getattr(ptr, insertion_attr)
                insertion_point[pointer_value.name.value] = value
            else:
            # Otherwise, just set the source key as insertion point:
                setattr(ptr, insertion_attr, value)

        return config


    def to_json(self):
        """
        Serialisera alla komponenter till rätt position i strukturen.
        """
        return self.as_model().model_dump()




from sqlmodel import SQLModel, Field
from typing import Optional, Dict
from enum import Enum

class ConnectionType(Enum):
    ssh = "ssh"
    telnet = "telnet"

class ManagementConnectionBase(SQLModel):
    primary: bool = True
    node_id: int
    connection_type: ConnectionType = Field(default=ConnectionType.ssh)
    target_ip: Optional[str] = None

class ManagementConnection(ManagementConnectionBase, table=True):
    __tablename__ = "mgmt_connection"
    id: int = Field(primary_key=True)

class ManagementConnectionResponse(ManagementConnection):
    id: int

class DeviceManagement(SQLModel):
    asset_id: int
    logical_node_id: int





from acex.configuration.components.base_component import ConfigComponent
# from acex.models.interfaces import PhysicalInterface, VirtualInterface, SubInterfaceAttributes
from acex.models.composed_configuration import (
    EthernetCsmacdInterface,
    Ieee8023adLagInterface,
    L3IpvlanInterface,
    SoftwareLoopbackInterface,
    SubInterface as SubInterfaceModel
)

from acex.models.composed_configuration import ReferenceFrom
from typing import Optional

class Interface(ConfigComponent):

    def _add_vrf(self):
        if self.kwargs.get('network_instance') is None:
            self.kwargs["network_instance"] = ReferenceFrom(pointer="network_instances.global.interfaces")
        else:
            network_instance = self.kwargs.pop("network_instance")
            self.kwargs["network_instance"] = ReferenceFrom(pointer=f"network_instances.{network_instance.name}.interfaces")



class Physical(Interface):
    type = "ethernetCsmacd"
    model_cls = EthernetCsmacdInterface

class LagInterface(Interface):
    """
    WIP :) 
    """
    ...

class Svi(Interface):
    type = "l3ipvlan"
    model_cls = L3IpvlanInterface

    def pre_init(self):
        referenced_vlan = self.kwargs.pop("vlan")
        self.kwargs["vlan_id"] = referenced_vlan.model.vlan_id.value
        self._add_vrf()



class Loopback(Interface):
    type = "softwareLoopback"
    model_cls = SoftwareLoopbackInterface

    def pre_init(self):
        self._add_vrf()

class Subinterface(Interface):
    type = "subinterface"
    model_cls = SubInterfaceModel

    def pre_init(self):
        vlan = self.kwargs.pop("vlan")
        self.kwargs["vlan"] = vlan.name 
        self.kwargs["vlan_id"] = vlan.model.vlan_id.value
        self._add_vrf()



from .asset import Asset, AssetCluster, AssetClusterLink, AssetResponse, Ned
from .logical_node import LogicalNode, LogicalNodeResponse, LogicalNodeListResponse
from .node import Node, NodeResponse
from .external_value import ExternalValue
from .attribute_value import AttributeValue
from .device_config import StoredDeviceConfig, DeviceConfig
from .management_connections import ManagementConnection, ManagementConnectionResponse, ManagementConnectionBase

system_models = [Asset, Ned, LogicalNode, Node]



from .sqlite import Sqlite
from .database import DatabasePlugin
from .in_memory import DatasourceInMemory
from .netbox import Netbox
from .integration_plugin_base import IntegrationPluginBase
from .integration_plugin_factory_base import IntegrationPluginFactoryBase


from .adapter_base import AdapterBase
from acex.models import Asset
from acex.models import ExternalValue
import json

class ProxyDatasource:
    """
    Proxy object to enable dotted syntax and query for 
    datasources within the plugin by referencing
    the resource name with ...{plugin}.data.{resource_name}({})
    """
    
    def __init__(self, plugin_adapter):
        self._plugin_adapter = plugin_adapter

    def __getattr__(self, resource_name: str):
        if resource_name not in self._plugin_adapter.plugin.DATA_TYPES:
            raise Exception(f"Resource type '{resource_name}' not defined in plugin: {self._plugin_adapter.plugin}")
        return lambda query: self._plugin_adapter._handle_datasource(resource_name, query)

class ProxyResource:
    """
    Proxy object to enable dotted syntax and query for 
    resources within the plugin by referencing
    the resource name with ...{plugin}.resource.{resource_name}({})
    """
    
    def __init__(self, plugin):
        self._plugin = plugin

    def __getattr__(self, resource_name: str):
        if resource_name not in self._plugin.RESOURCE_TYPES:
            raise Exception(f"Resource type '{resource_name}' not defined in plugin: {self._plugin}")
        return lambda query: self._plugin._handle_resource(resource_name, query)


class DatasourcePluginAdapter(AdapterBase):
    """
    Adapter säkerställer att rätt metodnamn används och
    inget annat. 

    Berikar respons med rätt responsetyper. 
    """

    def __init__(self, plugin):
        super().__init__(plugin)

    def __repr__(self):
        return f"d: {self.plugin}"


    @property
    def data(self):
        return ProxyDatasource(self)

    @property
    def resource(self):
        return ProxyResource(self)

    def _handle_datasource(self, resource_name: str, query: dict):
        data = self.query(resource_name, query)
        return data


    def _handle_resource(self, resource_name: str, query: dict):
        # TODO: Fixa create om inte state finns!
        data = self.plugin.query(resource_name, query)

        # IF NO STATE;
        # data = self.create()

        return "99.99.99.99/30"
        return data


    def create(self, asset: Asset): 
        if hasattr(self.plugin, "create"):
            return getattr(self.plugin, "create")(asset)
        raise NotImplementedError("Plugin does not support create operation")

    def get(self, kind: str, id: str): 
        if hasattr(self.plugin, "get"):
            return getattr(self.plugin, "get")(id)
        raise NotImplementedError("Plugin does not support get operation")

    def query(self, kind: str, filters: dict = None): 
        """
        The query method of the adapter will not actually 
        run the callable from the plugin, but construct an
        instance of an ExternalValue, to which all query parameters
        and details about the plugin and the actual Callable 
        is placed. The actual resolvement of the values are done 
        later by the configCompiler.
        """

        # Using the function from the very plugin
        func = getattr(self.plugin, "query")

        ev = ExternalValue(
            ev_type = "data",
            query=json.dumps(filters),
            value="_known after resolve_",
            kind=kind,
            plugin=str(self.plugin)
        )

        # Tilldela _callable efter objektskapande för PrivateAttr
        ev._callable = func
        return ev


    def update(self, asset: Asset): 
        if hasattr(self.plugin, "update"):
            return getattr(self.plugin, "update")(asset)
        raise NotImplementedError("Plugin does not support update operation")

    def delete(self, id: str): 
        if hasattr(self.plugin, "delete"):
            return getattr(self.plugin, "delete")(id)
        raise NotImplementedError("Plugin does not support delete operation")




from pydantic import BaseModel, Field
from typing import Optional, Dict, List, Literal, ClassVar
from enum import Enum

from acex.models.external_value import ExternalValue
from acex.models.attribute_value import AttributeValue
from acex.models.logging import (
    LoggingConfig,
    LoggingConsole,
    RemoteServer,
    LoggingEvents
)



class Metadata(BaseModel):
    type: str = "NoneType"

class Reference(BaseModel): ...

class ReferenceTo(Reference):
    pointer: str
    metadata: Optional[Dict] = {}

class ReferenceFrom(Reference):
    pointer: str
    metadata: Optional[Dict] = {}

class RenderedReference(BaseModel):
    from_ptr: str
    to_ptr: str

class SystemConfig(BaseModel):
    contact: Optional[AttributeValue[str]] = None
    domain_name: Optional[AttributeValue[str]] = None
    hostname: Optional[AttributeValue[str]] = None
    location: Optional[AttributeValue[str]] = None

class TripleA(BaseModel): ...


class Logging(BaseModel): 
    config: LoggingConfig = LoggingConfig()
    console: Optional[LoggingConsole] = None
    remote_servers: Optional[RemoteServer] = None
    events: Optional[LoggingEvents] = None


class NtpConfig(BaseModel):
    enabled: AttributeValue[bool] = False

class NtpServer(BaseModel):
    address: AttributeValue[str]
    port: Optional[AttributeValue[int]] = None
    version: Optional[AttributeValue[int]] = None
    association_typ: Optional[AttributeValue[str]] = None
    prefer: Optional[AttributeValue[bool]] = None
    source_interface: Optional[AttributeValue[str]] = None

class Ntp(BaseModel): 
    config: Optional[NtpConfig] = NtpConfig()
    servers: Optional[Dict[str, NtpServer]] = {}

class SshServer(BaseModel): 
    enable: Optional[AttributeValue[bool]] = None
    protocol_version: Optional[AttributeValue[int]] = 2
    timeout: Optional[AttributeValue[int]] = None
    auth_retries: Optional[AttributeValue[int]] = None
    source_interface: Optional[Reference] = None

class AuthorizedKeyAlgorithms(str, Enum):
    SSH_ED25519 = "ssh-ed25519"
    ECDSA_NISTP256 = "ecdsa-sha2-nistp256"
    ECDSA_NISTP384 = "ecdsa-sha2-nistp384"
    ECDSA_NISTP521 = "ecdsa-sha2-nistp521"
    RSA_SHA2_256 = "rsa-sha2-256"
    RSA_SHA2_512 = "rsa-sha2-512"
    SK_SSH_ED25519 = "sk-ssh-ed25519@openssh.com"
    SK_ECDSA_NISTP256 = "sk-ecdsa-sha2-nistp256@openssh.com"
    SSH_RSA = "ssh-rsa"
    SSH_DSS = "ssh-dss"

class AuthorizedKey(BaseModel):
    algorithm: AuthorizedKeyAlgorithms
    public_key: str

class Ssh(BaseModel): 
    config: Optional[SshServer] = None
    host_keys: Optional[Dict[str, AuthorizedKey]] = {}

class Acl(BaseModel): ...
class Lldp(BaseModel): ...

class Vlan(BaseModel):
    name: AttributeValue[str]
    vlan_id: Optional[AttributeValue[int]] = None
    vlan_name: Optional[AttributeValue[str]] = None
    network_instance: Optional[AttributeValue[str]] = None
    metadata: Optional[Metadata] = Metadata()

class Interface(BaseModel): 
    "Base class for all interfaces"
    index: AttributeValue[int]
    name: AttributeValue[str]

    description: Optional[AttributeValue[str]] = None
    enabled: Optional[AttributeValue[bool]] = None
    ipv4: Optional[AttributeValue[str]] = None
    
    metadata: Optional[Metadata] = Field(default_factory=Metadata)
    type: Literal[
        "ethernetCsmacd",
        "ieee8023adLag",
        "l3ipvlan",
        "softwareLoopback",
        "subinterface",
        ] = "ethernetCsmacd"
    
    model_config = {
        "discriminator": "type"
    }

class EthernetCsmacdInterface(Interface):
    "Physical Interface"
    type: Literal["ethernetCsmacd"] = "ethernetCsmacd"

    # Egenskaper för fysiska interface
    subinterfaces: list["SubInterface"] = Field(default_factory=list)
    speed: Optional[AttributeValue[int]] = None
    duplex: Optional[AttributeValue[str]] = None
    switchport_mode: Optional[AttributeValue[Literal["access", "trunk"]]] = None
    trunk_allowed_vlans: Optional[AttributeValue[List[int]]] = None
    native_vlan: Optional[AttributeValue[int]] = None
    access_vlan: Optional[AttributeValue[int]] = None
    vlan_id: Optional[AttributeValue[int]] = None
    voice_vlan: Optional[AttributeValue[int]] = None

class Ieee8023adLagInterface(Interface):
    "LAG Interface"
    type: Literal["ieee8023adLag"] = "ieee8023adLag"
    members: list[str] = Field(default_factory=list)

class L3IpvlanInterface(Interface):
    "SVI Interface"
    type: Literal["l3ipvlan"] = "l3ipvlan"
    vlan_id: Optional[int] = None

class SoftwareLoopbackInterface(Interface):
    "Loopback Interface"
    type: Literal["softwareLoopback"] = "softwareLoopback"

    # Loopback har varken vlan, duplex eller speed
    vlan_id: Optional[int] = None
    ipv4: Optional[AttributeValue[str]] = None

class SubInterface(Interface):
    "Subinterface"
    type: Literal["subinterface"] = "subinterface"

    vlan_id: Optional[int] = None
    ipv4: Optional[AttributeValue[str]] = None


class RouteTarget(BaseModel):
    value: str # TODO: Add constraints and validators... 

class ImportExportPolicy(BaseModel):
    export_route_target: Optional[List[RouteTarget]] = None
    import_route_target: Optional[List[RouteTarget]] = None

class InterInstancePolicy(BaseModel):
    import_export_policy: ImportExportPolicy

class NetworkInstance(BaseModel): 
    name: AttributeValue[str]
    description: Optional[AttributeValue[str]] = None
    vlans: Optional[Dict[str, Vlan]] = {}
    interfaces: Optional[Dict[str, Interface]] = {}
    inter_instance_policies: Optional[Dict[str, InterInstancePolicy]] = {}


class System(BaseModel):
    config: SystemConfig = SystemConfig()
    aaa: Optional[TripleA] = TripleA()
    logging: Optional[Logging] = Logging()
    ntp: Optional[Ntp] = Ntp()
    ssh: Optional[Ssh] = Ssh()

class ComposedConfiguration(BaseModel):
    system: Optional[System] = System()
    acl: Optional[Acl] = Acl()
    lldp: Optional[Lldp] = Lldp()
    interfaces: Dict[str, Interface] = {}
    network_instances: Dict[str, NetworkInstance] = {"global": NetworkInstance(name="global")}



from acex.configuration.components.base_component import ConfigComponent
from acex.models.composed_configuration import NtpServer as NtpServerAttributes

class NtpServer(ConfigComponent):
    type = "ntp_server"
    model_cls = NtpServerAttributes


import requests
import urllib3
from urllib3.exceptions import InsecureRequestWarning

# Stäng av SSL verification warnings
urllib3.disable_warnings(InsecureRequestWarning)


class RestClient:
    def __init__(self, base_url: str, verify_ssl: bool = True):
        self.base_url = base_url
        self.headers = {}
        self.verify_ssl = verify_ssl

    def _create_request_url(self, endpoint: str):
        return f"{self.base_url}{endpoint}"

    def request(self, method: str, endpoint: str, data=None, params=None):
        print(f"restclient: {method.upper()} {self.base_url}{endpoint}")
        url = self._create_request_url(endpoint)
        response = requests.request(method, url, headers=self.headers, json=data, params=params, verify=self.verify_ssl)
        return response

    def add_header(self, key: str, value: str):
        self.headers[key] = value

    def get(self, endpoint: str, params=None):
        return self.request("GET", endpoint, params=params)

    def post(self, endpoint: str, data=None, json=None):
        return self.request("POST", endpoint, data=data, json=json)

    def put(self, endpoint: str, data=None):
        # Implementation of PUT request
        pass

    def delete(self, endpoint: str):
        # Implementation of DELETE request
        pass


import importlib
import pkgutil
from typing import Dict, Type
from importlib.metadata import entry_points
from acex.plugins.neds.core import NetworkElementDriver

class NEDManager:
    def __init__(self):

        self.drivers: Dict[str, list[NetworkElementDriver]] = {}
        self.load_internal_drivers()
        self.load_external_drivers()

    def load_internal_drivers(self):
        """Ladda interna drivrutiner från plugins.neds.internal-katalogen."""
        base_path = "acex.plugins.neds.internal"
        for importer, module_name, _ in pkgutil.walk_packages(
            path=importlib.import_module(base_path).__path__,
            prefix=f"{base_path}."
        ):
            try:
                module = importlib.import_module(module_name)
                for attr_name in dir(module):
                    klass = getattr(module, attr_name)
                    if (
                        isinstance(klass, type)
                        and issubclass(klass, NetworkElementDriver)
                        and klass is not NetworkElementDriver
                    ):
                        instance = klass()
                        key = instance.key
                        self.drivers.setdefault(key, []).append(instance)
            except Exception as e:
                print(f"Fel vid laddning av {module_name}: {e}")


    def load_external_drivers(self):
        """Ladda externa drivrutiner via entry_points."""

        for entry_point in entry_points(group="acex.neds"):
            try:
                klass = entry_point.load()
                if issubclass(klass, NetworkElementDriver):
                    instance = klass()
                    key = instance.key
                    self.drivers.setdefault(key, []).append(instance)
            except Exception as e:
                print(f"Fel vid laddning av {entry_point.name}: {e}")

    def get_driver(self, driver_name: str, version: str = None) -> NetworkElementDriver:
        """Hämta en drivrutinsinstans efter namn och ev. version. Om version inte anges, returnera den med högst version."""
        candidates = self.drivers.get(driver_name.lower(), [])
        if not candidates:
            raise ValueError(f"Drivrutinen {driver_name} finns inte")
        if version:
            for driver in candidates:
                driver_version = getattr(type(driver), "version", None)
                if driver_version == version:
                    return driver
            raise ValueError(f"Drivrutinen {driver_name} med version {version} finns inte")
        if len(candidates) == 1:
            return candidates[0]
        try:
            from packaging.version import parse as parse_version
            candidates.sort(key=lambda d: parse_version(getattr(type(d), "version", "0.0.0")), reverse=True)
        except ImportError:
            candidates.sort(key=lambda d: getattr(type(d), "version", "0.0.0"), reverse=True)
        return candidates[0]

    def list_drivers(self) -> list[dict]:
        """Returnera en lista över tillgängliga drivrutinsnamn."""
        result = []
        for key, drivers in self.drivers.items():
            for driver in drivers:
                kind = type(driver)
                info = {
                    "name": getattr(kind, "name", key),
                    "class": kind.__name__,
                    "version": getattr(kind, "version", "n/a"),
                    "description": kind.__doc__ or "n/a",
                    "identifier": driver.key
                }
                result.append(info)
        return result


import inspect
import functools
from fastapi import APIRouter
from fastapi.responses import PlainTextResponse
from typing import Callable, Any, Optional

from acex.constants import BASE_URL


def get_response_model(func):
    sig = inspect.signature(func)
    return_annotation = sig.return_annotation
    if return_annotation is inspect.Signature.empty:
        return None
    return return_annotation

def get_request_model(func: Callable) -> Optional[type]:
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)

    for name, param in sig.parameters.items():
        if name in ("self", "cls"):
            continue
        param_type = type_hints.get(name)
        if param_type and isinstance(param_type, type) and issubclass(param_type, BaseModel):
            return param_type
    return None


def create_router(automation_engine):
    router = APIRouter(prefix=f"{BASE_URL}/inventory")
    tags = ["Inventory"]

    plug = getattr(automation_engine.inventory, "node_instances")
    for cap in plug.capabilities:
        func = getattr(plug, cap)
        path = plug.path(cap)
        method = plug.http_verb(cap)
        response_model = get_response_model(func)

        router.add_api_route(
            f"/node_instances{path}",
            func,
            methods=[method],
            response_model=response_model,
            tags=tags
        )

    router.add_api_route(
        "/node_instances/{id}/config",
        automation_engine.inventory.node_instances.get_rendered_config,
        methods=["GET"],
        response_class=PlainTextResponse,
        tags=tags
    )

    return router









from acex.models.asset import AssetClusterCreate, AssetCluster, AssetClusterLink
from fastapi import HTTPException

class AssetClusterManager:

    def __init__(self, db_manager):
        self.db = db_manager

    def create_cluster(self, payload: AssetClusterCreate) -> AssetCluster:
        session = next(self.db.get_session())
        try:
            cluster = AssetCluster.model_validate(payload)
            session.add(cluster)
            session.commit()
            session.refresh(cluster)
            return cluster
        finally:
            session.close()

    def list_clusters(self) -> list[AssetCluster]:
        session = next(self.db.get_session())
        try:
            return session.query(AssetCluster).all()
        finally:
            session.close()

    def get_cluster(self, id: int) -> dict:
        session = next(self.db.get_session())
        try:
            cluster = session.get(AssetCluster, id)
            if not cluster:
                raise HTTPException(status_code=404, detail="AssetCluster not found")
            # assets laddas automatiskt via relationship, men vi serialiserar till dict
            return {
                "id": cluster.id,
                "name": cluster.name,
                "assets": [
                    {
                        "id": asset.id,
                        "vendor": asset.vendor,
                        "serial_number": asset.serial_number,
                        "os": asset.os,
                        "os_version": asset.os_version,
                        "hardware_model": asset.hardware_model,
                        "ned_id": asset.ned_id
                    }
                    for asset in cluster.assets
                ]
            }
        finally:
            session.close()

    def update_cluster_assets(self, cluster_id: int, asset_ids: list[int]) -> AssetCluster:
        session = next(self.db.get_session())
        try:
            cluster = session.get(AssetCluster, cluster_id)
            if not cluster:
                raise HTTPException(status_code=404, detail="AssetCluster not found")

            # Ta bort gamla länkar
            session.exec(
                AssetClusterLink.__table__.delete().where(AssetClusterLink.cluster_id == cluster_id)
            )

            # Lägg till nya länkar
            for order, asset_id in enumerate(asset_ids):
                link = AssetClusterLink(asset_id=asset_id, cluster_id=cluster_id, order=order)
                session.add(link)

            session.commit()
            session.refresh(cluster)
            return cluster
        finally:
            session.close()

    def delete_cluster(self, id: int) -> None:
        session = next(self.db.get_session())
        try:
            cluster = session.get(AssetCluster, id)
            if not cluster:
                raise HTTPException(status_code=404, detail="AssetCluster not found")
            session.delete(cluster)
            session.commit()
        finally:
            session.close()


from acex.configuration.components.base_component import ConfigComponent
from acex.models.logging import (
    LoggingConfig,
    LoggingConsole,
    RemoteServer,
    LoggingEvent
)


class LoggingConfig(BaseModel):
    type= 'logging_config'
    model_cls = LoggingConfig


class LoggingConsole(BaseModel):
    type = 'console'
    model_cls = LoggingConsole


class LoggingServer(BaseModel):
    type = 'logging_server'
    model_cls = RemoteServer

class LoggingEvent(BaseModel):
    type = 'logging_event'
    model_cls = LoggingEvents



from .adapter_base import AdapterBase
from acex.models import Node, NodeResponse


class NodeAdapter(AdapterBase):

    def create(self, node: Node): 
        if hasattr(self.plugin, "create"):
            return getattr(self.plugin, "create")(node)

    def get(self, id: str) -> NodeResponse: 
        if hasattr(self.plugin, "get"):
            return getattr(self.plugin, "get")(id)

    def query(self, filters: dict = None) -> list[Node]: 
        if hasattr(self.plugin, "query"):
            print(filters)
            return getattr(self.plugin, "query")(filters)

    def update(self, id: str, node: Node): 
        if hasattr(self.plugin, "update"):
            return getattr(self.plugin, "update")(id, node)

    def delete(self, id: str): 
        if hasattr(self.plugin, "delete"):
            return getattr(self.plugin, "delete")(id)




from fastapi import APIRouter
from acex.constants import BASE_URL
from acex.plugins.neds.manager import NEDManager

nm = NEDManager()

def list_neds():
    neds = nm.list_drivers()
    return neds

def get_ned(driver_name: str, version: str = ""):
    ned = nm.get_driver(driver_name)

    return ned

def create_router(automation_engine):
    router = APIRouter(prefix=f"{BASE_URL}/neds")
    tags = ["Inventory"]
    router.add_api_route(
        "/",
        list_neds,
        methods=["GET"],
        tags=tags
    )
    router.add_api_route(
        "/{ned_id}",
        get_ned,
        methods=["GET"],
        tags=tags
    )

    return router





from .integration_plugin_base import IntegrationPluginBase
from .integration_plugin_factory_base import IntegrationPluginFactoryBase
from acex.utils import RestClient
from acex.models import AssetResponse, LogicalNodeResponse

import json
from requests import Response


# TODO: Fixa en egen hantering av response formatering. 
# Behöver få med:
# - DatasourceValue
# - Metadata
# - Modellering och validering av standardobjekt
# - Hämta interface uppmappade i netbox för assets? 


class Netbox(IntegrationPluginFactoryBase): 

    def __init__(self, url: str, token: str, verify_ssl: bool = True):
        """
        Used to define the connection to netbox. This is a factory class
        that will create a NetboxPluginInstance for each model used.
        """
        self.base_url = f"{url}api/"
        self.token = token
        self.rest = RestClient(self.base_url, verify_ssl=verify_ssl)
        self.rest.add_header("Authorization", f"Token {self.token}")
        self.rest.add_header("Content-Type", "application/json")


    def create_plugin(self, type: str|None = None) -> 'NetboxPlugin':
        """
        Create a plugin instance for a specific model.
        If no type is selected, generic plugin is returned.
        :param type: string representing the model/type to retreive.
        :return: An instance of NetboxPlugin.
        """
        if type is not None:
            return NetboxMockPlugin(type, restclient=self.rest)
        else:
            return NetboxMockPlugin(restclient=self.rest)


class NetboxMockPlugin(IntegrationPluginBase):
    RESOURCE_TYPES = [
        "prefixes",
        "ip_addresses"
    ]

    DATA_TYPES= [
        "ip_addresses"
    ]

    def __init__(self, restclient: RestClient):
        self.rest = restclient
    
    def __repr__(self):
        return f"{self.__class__.__name__}:{self.rest.base_url}"

    def _type_to_endpoint(self, kind) -> str:
        """
        Maps the endpoint for a given object kind
        """
        match kind:
            case "ip_addresses":
                return "ipam/ip-addresses"

    def _get_output_from_response(self, kind, json_dict: dict):
        """
        Extract data from a json response based on object type.
        """
        match kind:
            case "ip_addresses":
                return json_dict["address"]

    def query(self, kind: str, query: dict): 
        return "192.0.2.1/24"





from acex.configuration.components.base_component import ConfigComponent
from acex.configuration.components.interfaces import Interface
from acex.configuration.components.vlan import Vlan

from acex.models.composed_configuration import (
    NetworkInstance
)


class L3Vrf(ConfigComponent): 
    type = "l3vrf"
    model_cls = NetworkInstance


class L2Domain(ConfigComponent):
    type = "l2vsi"
    model_cls = NetworkInstance

